from .common import DtoObject


class LeaguesListDto(DtoObject):
    pass


class LeagueListDto(DtoObject):
    pass


class LeagueItemDto(DtoObject):
    pass


class MiniSeriesDto(DtoObject):
    pass


class LeaguePositionsDto(DtoObject):
    pass


class ChallengerLeagueListDto(DtoObject):
    pass


class MasterLeagueListDto(DtoObject):
    pass
