"""
This file defines constants which
are used in the various tests to
validate the API functions.
"""


# Summoner name to be used for Standard API calls, defaulting to NA
SUMMONER_NAME = "Kalturi"

# Unknown Summoner, used to validate that an error is thrown
UNKNOWN_SUMMONER_NAME = "abcaklsjdlakakdjlsakjdsdjlaksjdla"

# Champion name used for validating champion-related endpoints
CHAMP_NAME = "Karma"
